import json
import os
import gzip
import unittest
import networkx as nw
import numpy as np
import datetime
from copy import deepcopy

from openqaoa import QAOA, RQAOA
from openqaoa.problems import NumberPartition
from openqaoa.algorithms import QAOAResult, RQAOAResult
from openqaoa.algorithms.baseworkflow import Workflow
from openqaoa.utilities import (
    X_mixer_hamiltonian,
    XY_mixer_hamiltonian,
    is_valid_uuid,
    ground_state_hamiltonian,
)
from openqaoa.algorithms.workflow_properties import (
    BackendProperties,
    ClassicalOptimizer,
    CircuitProperties,
)
from openqaoa.algorithms.rqaoa.rqaoa_workflow_properties import RqaoaParameters
from openqaoa.backends import create_device, DeviceLocal
from openqaoa.backends.cost_function import cost_function

from openqaoa.backends.devices_core import SUPPORTED_LOCAL_SIMULATORS
from openqaoa.qaoa_components import (
    Hamiltonian,
    QAOADescriptor,
    QAOAVariationalStandardParams,
    QAOAVariationalStandardWithBiasParams,
    QAOAVariationalExtendedParams,
    QAOAVariationalFourierParams,
    QAOAVariationalFourierExtendedParams,
    QAOAVariationalFourierWithBiasParams,
)
from openqaoa.backends import QAOAvectorizedBackendSimulator
from openqaoa.backends.basebackend import QAOABaseBackendStatevector
from openqaoa.problems import MinimumVertexCover, QUBO, MaximumCut
from openqaoa.optimizers.qaoa_optimizer import available_optimizers
from openqaoa.optimizers.training_vqa import (
    ScipyOptimizer,
    CustomScipyGradientOptimizer,
    PennyLaneOptimizer,
)

from openqaoa.qaoa_components.variational_parameters.variational_params_factory import (
    PARAMS_CLASSES_MAPPER,
)


def _compare_qaoa_results(dict_old, dict_new):
    for key in dict_old.keys():
        if key == "cost_hamiltonian":  # CHECK WHAT DO WITH THIS
            pass
        elif key == "_QAOAResult__type_backend":
            if issubclass(dict_old[key], QAOABaseBackendStatevector):
                assert (
                    dict_new[key] == QAOABaseBackendStatevector
                ), "Type of backend is not correct."
            else:
                assert dict_new[key] == "", "Type of backend should be empty string."
        elif key == "optimized":
            for key2 in dict_old[key].keys():
                if key2 == "measurement_outcomes":
                    assert np.all(
                        dict_old[key][key2] == dict_new[key][key2]
                    ), "Optimized params are not the same."
                else:
                    assert (
                        dict_old[key][key2] == dict_new[key][key2]
                    ), "Optimized params are not the same."
        elif key == "intermediate":
            for key2 in dict_old[key].keys():
                if key2 == "measurement_outcomes":
                    for step in range(len(dict_old[key][key2])):
                        assert np.all(
                            dict_old[key][key2][step] == dict_new[key][key2][step]
                        ), f"Intermediate params are not the same. Expected {dict_old[key][key2][step]} but \
                            received {dict_new[key][key2][step]}"
                else:
                    assert (
                        dict_old[key][key2] == dict_new[key][key2]
                    ), f"Intermediate params are not the same. Expected {dict_old[key][key2]}, but \
                        received {dict_new[key][key2]}"
        else:
            assert dict_old[key] == dict_new[key], f"'{key}' is not the same"


def _test_keys_in_dict(obj, expected_keys):
    """
    private function to test the keys.
    It recursively tests the keys of the nested dictionaries, or lists of dictionaries
    """

    if isinstance(obj, dict):
        for key in obj:
            if key in expected_keys.keys():
                expected_keys[key] = True

            if isinstance(obj[key], dict):
                _test_keys_in_dict(obj[key], expected_keys)
            elif isinstance(obj[key], list):
                for item in obj[key]:
                    _test_keys_in_dict(item, expected_keys)
    elif isinstance(obj, list):
        for item in obj:
            _test_keys_in_dict(item, expected_keys)


class TestingVanillaQAOA(unittest.TestCase):

    """
    Unit test based testing of the QAOA workflow class
    """

    def test_vanilla_qaoa_default_values(self):
        q = QAOA()
        assert q.circuit_properties.p == 1
        assert q.circuit_properties.param_type == "standard"
        assert q.circuit_properties.init_type == "ramp"
        assert q.device.device_location == "local"
        assert q.device.device_name == "vectorized"

    def test_end_to_end_vectorized(self):
        g = nw.circulant_graph(6, [1])
        vc = MinimumVertexCover(g, field=1.0, penalty=10).qubo

        q = QAOA()
        q.set_classical_optimizer(optimization_progress=True)
        q.compile(vc)
        q.optimize()

        result = q.result.most_probable_states["solutions_bitstrings"][0]
        assert "010101" == result or "101010" == result

    def test_set_device_local(self):
        """ "
        Check that all local devices are correctly initialised
        """
        q = QAOA()
        for d in q.local_simulators:
            q.set_device(create_device(location="local", name=d))
            assert type(q.device) == DeviceLocal
            assert q.device.device_name == d
            assert q.device.device_location == "local"

    def test_compile_before_optimise(self):
        """
        Assert that compilation has to be called before optimisation
        """

        q = QAOA()
        q.set_classical_optimizer(optimization_progress=True)

        self.assertRaises(ValueError, lambda: q.optimize())

    def test_cost_hamil(self):
        g = nw.circulant_graph(6, [1])
        problem = MinimumVertexCover(g, field=1.0, penalty=10)
        qubo_problem = problem.qubo

        test_hamil = Hamiltonian.classical_hamiltonian(
            terms=qubo_problem.terms,
            coeffs=qubo_problem.weights,
            constant=qubo_problem.constant,
        )

        q = QAOA()

        q.compile(problem=qubo_problem)

        self.assertEqual(q.cost_hamil.expression, test_hamil.expression)
        self.assertEqual(
            q.qaoa_descriptor.cost_hamiltonian.expression, test_hamil.expression
        )

    def test_qaoa_brute_force(self):
        """
        test the brute_force method
        """
        # problem
        problem = MinimumVertexCover.random_instance(
            n_nodes=6, edge_probability=0.8
        ).qubo

        # check if the brute force method gets the same results as the ground state hamiltonian method
        qaoa = QAOA()
        qaoa.compile(problem)
        qaoa.solve_brute_force()
        min_energy_bf, config_strings_bf = (
            qaoa.brute_force_results["energy"],
            qaoa.brute_force_results["configuration"],
        )
        (min_energy_gsh, config_strings_ghs) = ground_state_hamiltonian(qaoa.cost_hamil)
        assert (
            min_energy_bf == min_energy_gsh and config_strings_bf == config_strings_ghs
        ), f"The energy and config strings of brute forcing should match the ground state hamiltonian method results"

        # Check if an uncompiled QAOA raises an error when calling its brute_force method
        qaoa_uncompiled = QAOA()
        error = False
        try:
            qaoa_uncompiled.solve_brute_force()
        except Exception:
            error = True
        assert error, f"An uncompiled QAOA should raise an error when brute-forcing it"

        # Check if bounded=True disallows computation for more than 25 qubits
        qaoa = QAOA()
        large_problem = MaximumCut.random_instance(n_nodes=26, edge_probability=1).qubo
        qaoa.compile(large_problem)
        error = False
        try:
            qaoa.solve_brute_force()
        except:
            error = True
        assert (
            error
        ), f"Brute forcing should not compute for large problems (> 25 qubits) when bounded=True"

    def test_set_circuit_properties_fourier_q(self):
        """
        The value of q should be None if the param_type used is not fourier.
        Else if param_type is fourier, fourier_extended or fourier_w_bias, it
        should be the value of q, if it is provided.
        """

        fourier_param_types = ["fourier", "fourier_extended", "fourier_w_bias"]

        q = QAOA()

        for each_param_type in fourier_param_types:
            q.set_circuit_properties(param_type=each_param_type, q=1)
            self.assertEqual(q.circuit_properties.q, 1)

        q.set_circuit_properties(param_type="standard", q=1)

        self.assertEqual(q.circuit_properties.q, None)

    def test_set_circuit_properties_annealing_time_linear_ramp_time(self):
        """
        Check that linear_ramp_time and annealing_time are updated appropriately
        as the value of p is changed.
        """

        q = QAOA()

        q.set_circuit_properties(p=3)

        self.assertEqual(q.circuit_properties.annealing_time, 0.7 * 3)
        self.assertEqual(q.circuit_properties.linear_ramp_time, 0.7 * 3)

        q.set_circuit_properties(p=2)

        self.assertEqual(q.circuit_properties.annealing_time, 0.7 * 2)
        self.assertEqual(q.circuit_properties.linear_ramp_time, 0.7 * 2)

    def test_set_circuit_properties_qaoa_descriptor_mixer_x(self):
        """
        Checks if the X mixer created by the X_mixer_hamiltonian method
        and the automated methods in workflows do the same thing.

        For each qubit, there should be 1 RXGateMap per layer of p.
        """

        nodes = 6
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)

        q = QAOA()
        q.set_circuit_properties(mixer_hamiltonian="x", p=2)

        q.compile(problem=problem.qubo)

        self.assertEqual(type(q.qaoa_descriptor), QAOADescriptor)
        self.assertEqual(q.qaoa_descriptor.p, 2)

        mixer_hamil = X_mixer_hamiltonian(n_qubits=nodes)

        self.assertEqual(q.mixer_hamil.expression, mixer_hamil.expression)

        self.assertEqual(len(q.qaoa_descriptor.mixer_qubits_singles), 6)
        self.assertEqual(len(q.qaoa_descriptor.mixer_qubits_pairs), 0)
        for each_gatemap_name in q.qaoa_descriptor.mixer_qubits_singles:
            self.assertEqual(each_gatemap_name, "RXGateMap")

        for j in range(2):
            for i in range(6):
                self.assertEqual(q.qaoa_descriptor.mixer_blocks[j][i].qubit_1, i)

    def test_set_circuit_properties_qaoa_descriptor_mixer_xy(self):
        """
        Checks if the XY mixer created by the XY_mixer_hamiltonian method
        and the automated methods in workflows do the same thing.

        Depending on the qubit connectivity selected. (chain, full or star)
        For each pair of connected qubits, there should be 1 RXXGateMap and RYYGateMap per layer of p.
        """

        g_c = nw.circulant_graph(6, [1])
        g_f = nw.complete_graph(6)
        # A 5-sided star graoh requires 6 qubit. (Center Qubit of the pattern)
        g_s = nw.star_graph(5)
        problems = [
            MinimumVertexCover(g_c, field=1.0, penalty=10),
            MinimumVertexCover(g_f, field=1.0, penalty=10),
            MinimumVertexCover(g_s, field=1.0, penalty=10),
        ]
        qubit_connectivity_name = ["chain", "full", "star"]

        for i in range(3):
            q = QAOA()
            q.set_circuit_properties(
                mixer_hamiltonian="xy",
                mixer_qubit_connectivity=qubit_connectivity_name[i],
                p=2,
            )

            q.compile(problem=problems[i].qubo)

            self.assertEqual(type(q.qaoa_descriptor), QAOADescriptor)
            self.assertEqual(q.qaoa_descriptor.p, 2)

            mixer_hamil = XY_mixer_hamiltonian(
                n_qubits=6, qubit_connectivity=qubit_connectivity_name[i]
            )

            self.assertEqual(q.mixer_hamil.expression, mixer_hamil.expression)

            self.assertEqual(len(q.qaoa_descriptor.mixer_qubits_singles), 0)
            for i in range(0, len(q.qaoa_descriptor.mixer_qubits_pairs), 2):
                self.assertEqual(q.qaoa_descriptor.mixer_qubits_pairs[i], "RXXGateMap")
                self.assertEqual(
                    q.qaoa_descriptor.mixer_qubits_pairs[i + 1], "RYYGateMap"
                )

    def test_set_circuit_properties_variate_params(self):
        """
        Ensure that the Varitional Parameter Object created based on the input string , param_type, is correct.

        TODO: Check if q=None is the appropriate default.
        """

        param_type_names = [
            "standard",
            "standard_w_bias",
            "extended",
            "fourier",
            "fourier_extended",
            "fourier_w_bias",
        ]
        object_types = [
            QAOAVariationalStandardParams,
            QAOAVariationalStandardWithBiasParams,
            QAOAVariationalExtendedParams,
            QAOAVariationalFourierParams,
            QAOAVariationalFourierExtendedParams,
            QAOAVariationalFourierWithBiasParams,
        ]

        nodes = 6
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)

        for i in range(len(object_types)):
            q = QAOA()
            q.set_circuit_properties(param_type=param_type_names[i], q=1)

            q.compile(problem=problem.qubo)

            self.assertEqual(type(q.variate_params), object_types[i])

    def test_set_circuit_properties_change(self):
        """
        Ensure that once a property has beefn changed via set_circuit_properties.
        The attribute has been appropriately updated.
        Updating all attributes at the same time.
        """

        #         default_pairings = {'param_type': 'standard',
        #                             'init_type': 'ramp',
        #                             'qubit_register': [],
        #                             'p': 1,
        #                             'q': None,
        #                             'annealing_time': 0.7,
        #                             'linear_ramp_time': 0.7,
        #                             'variational_params_dict': {},
        #                             'mixer_hamiltonian': 'x',
        #                             'mixer_qubit_connectivity': None,
        #                             'mixer_coeffs': None,
        #                             'seed': None}

        q = QAOA()

        # TODO: Some weird error related to the initialisation of QAOA here
        #         for each_key, each_value in default_pairings.items():
        #             print(each_key, getattr(q.circuit_properties, each_key), each_value)
        #             self.assertEqual(getattr(q.circuit_properties, each_key), each_value)

        update_pairings = {
            "param_type": "fourier",
            "init_type": "rand",
            "qubit_register": [0, 1],
            "p": 2,
            "q": 2,
            "annealing_time": 1.0,
            "linear_ramp_time": 1.0,
            "variational_params_dict": {"key": "value"},
            "mixer_hamiltonian": "xy",
            "mixer_qubit_connectivity": "chain",
            "mixer_coeffs": [0.1, 0.2],
            "seed": 45,
        }

        q.set_circuit_properties(**update_pairings)

        for each_key, each_value in update_pairings.items():
            self.assertEqual(getattr(q.circuit_properties, each_key), each_value)

    def test_set_circuit_properties_rejected_values(self):
        """
        Some properties of CircuitProperties Object return a ValueError
        if the specified property has not been whitelisted in the code.
        This checks that the ValueError is raised if the argument is not whitelisted.
        """

        q = QAOA()

        self.assertRaises(
            ValueError, lambda: q.set_circuit_properties(param_type="wrong name")
        )
        self.assertRaises(
            ValueError, lambda: q.set_circuit_properties(init_type="wrong name")
        )
        self.assertRaises(
            ValueError, lambda: q.set_circuit_properties(mixer_hamiltonian="wrong name")
        )
        self.assertRaises(ValueError, lambda: q.set_circuit_properties(p=-1))

    def test_set_backend_properties_change(self):
        """
        Ensure that once a property has been changed via set_backend_properties.
        The attribute has been appropriately updated.
        Updating all attributes at the same time.
        """

        default_pairings = {
            "prepend_state": None,
            "append_state": None,
            "init_hadamard": True,
            "n_shots": 100,
            "cvar_alpha": 1.0,
        }

        q = QAOA()

        for each_key, each_value in default_pairings.items():
            self.assertEqual(getattr(q.backend_properties, each_key), each_value)

        update_pairings = {
            "prepend_state": [[0, 0]],
            "append_state": [[0, 0]],
            "init_hadamard": False,
            "n_shots": 10,
            "cvar_alpha": 0.5,
        }

        q.set_backend_properties(**update_pairings)

        for each_key, each_value in update_pairings.items():
            self.assertEqual(getattr(q.backend_properties, each_key), each_value)

    def test_set_backend_properties_check_backend_vectorized(self):
        """
        Check if the backend returned by set_backend_properties is correct
        Based on the input device.
        Also Checks if defaults from workflows are used in the backend.
        """

        nodes = 6
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)
        qubo_problem = problem.qubo

        q = QAOA()
        q.set_device(create_device(location="local", name="vectorized"))
        q.compile(problem=qubo_problem)

        self.assertEqual(type(q.backend), QAOAvectorizedBackendSimulator)

        self.assertEqual(q.backend.init_hadamard, True)
        self.assertEqual(q.backend.prepend_state, None)
        self.assertEqual(q.backend.append_state, None)
        self.assertEqual(q.backend.cvar_alpha, 1)

        self.assertRaises(AttributeError, lambda: q.backend.n_shots)

    def test_set_backend_properties_check_backend_vectorized_w_custom(self):
        """
        Check if the backend returned by set_backend_properties is correct
        Based on the input device.
        Uses custom values for attributes in backend_properties and checks if the
        backend object responds appropriately.
        """

        nodes = 3
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)
        qubo_problem = problem.qubo

        q = QAOA()
        q.set_device(create_device(location="local", name="vectorized"))

        prepend_state_rand = np.random.rand(2**3)
        append_state_rand = np.eye(2**3)

        update_pairings = {
            "prepend_state": prepend_state_rand,
            "append_state": append_state_rand,
            "init_hadamard": False,
            "n_shots": 10,
            "cvar_alpha": 1,
        }

        q.set_backend_properties(**update_pairings)

        q.compile(problem=qubo_problem)

        self.assertEqual(type(q.backend), QAOAvectorizedBackendSimulator)

        self.assertEqual(q.backend.init_hadamard, False)
        self.assertEqual((q.backend.prepend_state == prepend_state_rand).all(), True)
        self.assertEqual((q.backend.append_state == append_state_rand).all(), True)
        self.assertEqual(q.backend.cvar_alpha, 1)

        self.assertRaises(AttributeError, lambda: q.backend.n_shots)

    def test_set_backend_properties_check_backend_vectorized_error_values(self):
        """
        If the values provided from the workflows are incorrect, we should
        receive the appropriate error messages from the vectorized backend.

        Checks:
        Incorrect size of prepend state and append state.
        """

        nodes = 3
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)
        qubo_problem = problem.qubo

        q = QAOA()
        q.set_device(create_device(location="local", name="vectorized"))

        prepend_state_rand = np.random.rand(2**2)

        update_pairings = {"prepend_state": prepend_state_rand, "append_state": None}

        q.set_backend_properties(**update_pairings)

        self.assertRaises(ValueError, lambda: q.compile(problem=qubo_problem))

        q = QAOA()
        q.set_device(create_device(location="local", name="vectorized"))

        append_state_rand = np.random.rand(2**2, 2**2)

        update_pairings = {"prepend_state": None, "append_state": append_state_rand}

        q.set_backend_properties(**update_pairings)

        self.assertRaises(ValueError, lambda: q.compile(problem=qubo_problem))

    def test_set_classical_optimizer_defaults(self):
        """
        Check if the fields in the default classical_optimizer dict are correct
        """

        default_pairings = {
            "optimize": True,
            "method": "cobyla",
            "maxiter": 100,
            "jac": None,
            "hess": None,
            "constraints": None,
            "bounds": None,
            "tol": None,
            "optimizer_options": None,
            "jac_options": None,
            "hess_options": None,
            "optimization_progress": False,
            "cost_progress": True,
            "parameter_log": True,
        }

        q = QAOA()

        for each_key, each_value in default_pairings.items():
            self.assertEqual(getattr(q.classical_optimizer, each_key), each_value)

            # if each_value is None: LD --> I don't think we really need this test, since asdict()
            #     assert isinstance(q.classical_optimizer.each_key, None)

    def test_set_classical_optimizer_jac_hess_casing(self):
        """
        jac and hess should be in lower case if it is a string.
        """

        q = QAOA()
        q.set_classical_optimizer(jac="JaC", hess="HeSS")

        self.assertEqual(q.classical_optimizer.jac, "jac")
        self.assertEqual(q.classical_optimizer.hess, "hess")

    def test_set_classical_optimizer_method_selectors(self):
        """
        Different methods would return different Optimizer classes.
        Check that the correct class is returned.
        """

        nodes = 6
        edge_probability = 0.6
        g = nw.generators.fast_gnp_random_graph(n=nodes, p=edge_probability)
        problem = MinimumVertexCover(g, field=1.0, penalty=10)
        qubo_problem = problem.qubo

        for each_method in available_optimizers()["scipy"]:
            q = QAOA()
            q.set_classical_optimizer(method=each_method, jac="grad_spsa")
            q.compile(problem=qubo_problem)

            self.assertEqual(isinstance(q.optimizer, ScipyOptimizer), True)
            self.assertEqual(
                isinstance(q.optimizer, CustomScipyGradientOptimizer), False
            )
            self.assertEqual(isinstance(q.optimizer, PennyLaneOptimizer), False)

        for each_method in available_optimizers()["custom_scipy_gradient"]:
            q = QAOA()
            q.set_classical_optimizer(
                method=each_method, jac="grad_spsa", hess="finite_difference"
            )
            q.compile(problem=qubo_problem)

            self.assertEqual(isinstance(q.optimizer, ScipyOptimizer), False)
            self.assertEqual(
                isinstance(q.optimizer, CustomScipyGradientOptimizer), True
            )
            self.assertEqual(isinstance(q.optimizer, PennyLaneOptimizer), False)

        for each_method in available_optimizers()["custom_scipy_pennylane"]:
            q = QAOA()
            q.set_classical_optimizer(method=each_method, jac="grad_spsa")
            q.compile(problem=qubo_problem)

            self.assertEqual(isinstance(q.optimizer, ScipyOptimizer), False)
            self.assertEqual(
                isinstance(q.optimizer, CustomScipyGradientOptimizer), False
            )
            self.assertEqual(isinstance(q.optimizer, PennyLaneOptimizer), True)

    def test_set_header(self):
        """
        Test the test_set_header method of the QAOA class. Step by step it is checked that the header is set correctly.
        """

        # create a QAOA object
        qaoa: QAOA = QAOA()

        # check if the header values are set to None, except for the experiment_id and algorithm
        for key, value in qaoa.header.items():
            if key == "experiment_id":
                assert is_valid_uuid(
                    qaoa.header["experiment_id"]
                ), "The experiment_id is not a valid uuid."
            elif key == "algorithm":
                assert qaoa.header["algorithm"] == "qaoa"
            else:
                assert (
                    value is None
                ), "The value of the key {} (of the dictionary qaoa.header) is not None, when it should be.".format(
                    key
                )

        # save the experiment_id
        experiment_id = qaoa.header["experiment_id"]

        # set the header
        qaoa.set_header(
            project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
            description="test",
            run_by="OpenQAOA",
            provider="-",
            target="vectorized",
            cloud="local",
            client="-",
        )

        # check that the experiment_id has not changed, since it is not set in the set_header method
        assert (
            qaoa.header["experiment_id"] == experiment_id
        ), "The experiment_id has changed when it should not have."

        # now set the experiment_id
        experiment_id = experiment_id[:-2] + "00"

        # set the header
        qaoa.set_header(
            project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
            description="test",
            run_by="OpenQAOA",
            provider="-",
            target="vectorized",
            cloud="local",
            client="-",
            experiment_id=experiment_id,
        )

        # check if the header values are set to the correct values, except for the
        # qubit_number, atomic_id, execution_time_start, and execution_time_end (which are set to None)
        dict_values = {
            "experiment_id": experiment_id,
            "project_id": "8353185c-b175-4eda-9628-b4e58cb0e41b",
            "algorithm": "qaoa",
            "description": "test",
            "run_by": "OpenQAOA",
            "provider": "-",
            "target": "vectorized",
            "cloud": "local",
            "client": "-",
            "qubit_number": None,
            "atomic_id": None,
            "execution_time_start": None,
            "execution_time_end": None,
        }
        for key, value in qaoa.header.items():
            assert (
                dict_values[key] == value
            ), "The value of the key {} (of the dictionary qaoa.header) is not correct.".format(
                key
            )

        # compile the QAOA object
        qaoa.compile(problem=QUBO.random_instance(n=8))

        # check if the header values are still set to the correct values, except for execution_time_start, and
        # execution_time_end (which are set to None).
        # Now atomic_id should be set to a valid uuid.
        # And qubit_number should be set to 8 (number of qubits of the problem)
        dict_values["qubit_number"] = 8
        for key, value in qaoa.header.items():
            if key not in ["atomic_id"]:
                assert (
                    dict_values[key] == value
                ), "The value of the key {} (of the dictionary qaoa.header) is not correct.".format(
                    key
                )
        assert is_valid_uuid(
            qaoa.header["atomic_id"]
        ), "The atomic_id is not a valid uuid."

        # save the atomic_id
        atomic_id = qaoa.header["atomic_id"]

        # optimize the QAOA object
        qaoa.optimize()

        # check if the header values are still set to the correct values, now everything should be set to a valid value
        # (execution_time_start and execution_time_end should be integers>1672933928)
        dict_values["atomic_id"] = atomic_id
        for key, value in qaoa.header.items():
            if key not in ["execution_time_start", "execution_time_end"]:
                assert (
                    dict_values[key] == value
                ), "The value of the key {} (of the dictionary qaoa.header) is not correct.".format(
                    key
                )
        assert datetime.datetime.strptime(
            qaoa.header["execution_time_start"], "%Y-%m-%dT%H:%M:%S"
        ), "The execution_time_start is not valid."
        assert datetime.datetime.strptime(
            qaoa.header["execution_time_end"], "%Y-%m-%dT%H:%M:%S"
        ), "The execution_time_end is not valid."

        # test if an error is raised when the project_id is not a valid string
        error = False
        try:
            qaoa.set_header(
                project_id="test",
                description="test",
                run_by="OpenQAOA",
                provider="-",
                target="vectorized",
                cloud="local",
                client="-",
            )
        except:
            error = True
        assert error, "The project_id is not valid string, but no error was raised."

        # test if an error is raised when the experiment_id is not a valid string
        error = False
        try:
            qaoa.set_header(
                project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
                experiment_id="test",
                description="test",
                run_by="OpenQAOA",
                provider="-",
                target="vectorized",
                cloud="local",
                client="-",
            )
        except:
            error = True
        assert error, "The experiment_id is not valid string, but no error was raised."

    def test_set_exp_tags(self):
        """
        Test the set_exp_tags method of the QAOA class.
        """

        qaoa = QAOA()
        qaoa.set_exp_tags(tags={"tag1": "value1", "tag2": "value2"})
        qaoa.set_exp_tags(tags={"tag1": "value9"})
        qaoa.compile(problem=QUBO.random_instance(n=8))
        qaoa.optimize()

        tags = {
            "tag1": "value9",
            "tag2": "value2",
            "init_type": "ramp",
            "optimizer_method": "cobyla",
            "p": 1,
            "param_type": "standard",
            "qubit_number": 8,
        }

        assert qaoa.exp_tags == tags, "Experiment tags are not set correctly."

        error = False
        try:
            qaoa.set_exp_tags(tags={"tag1": complex(1, 2)})
        except:
            error = True
        assert error, "Experiment tag values should be primitives."

        error = False
        try:
            qaoa.set_exp_tags(tags={(1, 2): "test"})
        except:
            error = True
        assert error, "Experiment tag keys should be strings."

    def test_qaoa_asdict_dumps(self):
        """Test the asdict method of the QAOA class."""

        # qaoa
        qaoa = QAOA()
        qaoa.compile(problem=QUBO.random_instance(n=8))
        # set the header
        qaoa.set_header(
            project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
            description="test",
            run_by="OpenQAOA",
            provider="-",
            target="vectorized",
            cloud="local",
            client="-",
        )
        qaoa.optimize()

        # check QAOA asdict
        self.__test_expected_keys(qaoa.asdict(), method="asdict")

        # check QAOA asdict deleting some keys
        exclude_keys = ["corr_matrix", "number_steps"]
        self.__test_expected_keys(
            qaoa.asdict(exclude_keys=exclude_keys), exclude_keys, method="asdict"
        )

        # check QAOA dumps
        self.__test_expected_keys(json.loads(qaoa.dumps()), method="dumps")

        # check QAOA dumps deleting some keys
        exclude_keys = ["parent_id", "counter"]
        self.__test_expected_keys(
            json.loads(qaoa.dumps(exclude_keys=exclude_keys)),
            exclude_keys,
            method="dumps",
        )

        # check QAOA dump
        file_name = "test_dump_qaoa.json"
        project_id, experiment_id, atomic_id = (
            qaoa.header["project_id"],
            qaoa.header["experiment_id"],
            qaoa.header["atomic_id"],
        )
        full_name = f"{project_id}--{experiment_id}--{atomic_id}--{file_name}"

        qaoa.dump(file_name, indent=None, prepend_id=True)
        assert os.path.isfile(full_name), "Dump file does not exist"
        with open(full_name, "r") as file:
            assert file.read() == qaoa.dumps(
                indent=None
            ), "Dump file does not contain the correct data"
        os.remove(full_name)

        # check QAOA dump whitout prepending the experiment_id and atomic_id
        qaoa.dump(file_name, indent=None, prepend_id=False)
        assert os.path.isfile(
            file_name
        ), "Dump file does not exist, when not prepending the experiment_id and atomic_id"

        # check QAOA dump fails when the file already exists
        error = False
        try:
            qaoa.dump(file_name, indent=None, prepend_id=False)
        except FileExistsError:
            error = True
        assert error, "Dump file does not fail when the file already exists"

        # check that we can overwrite the file
        qaoa.dump(file_name, indent=None, prepend_id=False, overwrite=True)
        assert os.path.isfile(file_name), "Dump file does not exist, when overwriting"
        os.remove(file_name)

        # check QAOA dump fails when prepend_id is True and file_name is not given
        error = False
        try:
            qaoa.dump(prepend_id=False)
        except ValueError:
            error = True
        assert (
            error
        ), "Dump file does not fail when prepend_id is True and file_name is not given"

        # check QAOA dump with no arguments
        error = False
        try:
            qaoa.dump()
        except ValueError:
            error = True
        assert (
            error
        ), "Dump file does not fail when no arguments are given, should be the same as dump(prepend_id=False)"

        # check you can dump to a file with no arguments, just prepend_id=True
        qaoa.dump(prepend_id=True)
        assert os.path.isfile(
            f"{project_id}--{experiment_id}--{atomic_id}.json"
        ), "Dump file does not exist, when no name is given"
        os.remove(f"{project_id}--{experiment_id}--{atomic_id}.json")

        # check QAOA dump deleting some keys
        exclude_keys = ["schedule", "singlet"]
        qaoa.dump(file_name, exclude_keys=exclude_keys, indent=None, prepend_id=True)
        assert os.path.isfile(
            full_name
        ), "Dump file does not exist, when deleting some keys"
        with open(full_name, "r") as file:
            assert file.read() == qaoa.dumps(
                exclude_keys=exclude_keys, indent=None
            ), "Dump file does not contain the correct data, when deleting some keys"
        os.remove(full_name)

        # check QAOA dump with compression
        qaoa.dump(file_name, compresslevel=2, indent=None, prepend_id=True)
        assert os.path.isfile(
            full_name + ".gz"
        ), "Dump file does not exist, when compressing"
        with gzip.open(full_name + ".gz", "rb") as file:
            assert (
                file.read() == qaoa.dumps(indent=None).encode()
            ), "Dump file does not contain the correct data, when compressing"
        os.remove(full_name + ".gz")

    def __test_expected_keys(self, obj, exclude_keys=[], method="asdict"):
        """
        method to test if the dictionary has all the expected keys
        """

        # create a dictionary with all the expected keys and set them to False
        expected_keys = [
            "header",
            "atomic_id",
            "experiment_id",
            "project_id",
            "algorithm",
            "description",
            "run_by",
            "provider",
            "target",
            "cloud",
            "client",
            "qubit_number",
            "execution_time_start",
            "execution_time_end",
            "metadata",
            "problem_type",
            "n_shots",
            "optimizer_method",
            "param_type",
            "init_type",
            "p",
            "data",
            "exp_tags",
            "input_problem",
            "terms",
            "weights",
            "constant",
            "n",
            "problem_instance",
            "input_parameters",
            "device",
            "device_location",
            "device_name",
            "backend_properties",
            "init_hadamard",
            "prepend_state",
            "append_state",
            "cvar_alpha",
            "noise_model",
            "initial_qubit_mapping",
            "seed_simulator",
            "qiskit_simulation_method",
            "active_reset",
            "rewiring",
            "disable_qubit_rewiring",
            "classical_optimizer",
            "optimize",
            "method",
            "maxiter",
            "maxfev",
            "jac",
            "hess",
            "constraints",
            "bounds",
            "tol",
            "optimizer_options",
            "jac_options",
            "hess_options",
            "parameter_log",
            "optimization_progress",
            "cost_progress",
            "save_intermediate",
            "circuit_properties",
            "qubit_register",
            "q",
            "variational_params_dict",
            "total_annealing_time",
            "annealing_time",
            "linear_ramp_time",
            "mixer_hamiltonian",
            "mixer_qubit_connectivity",
            "mixer_coeffs",
            "seed",
            "result",
            "evals",
            "number_of_evals",
            "jac_evals",
            "qfim_evals",
            "most_probable_states",
            "solutions_bitstrings",
            "bitstring_energy",
            "intermediate",
            "angles",
            "cost",
            "measurement_outcomes",
            "job_id",
            "optimized",
            "eval_number",
        ]
        expected_keys = {item: False for item in expected_keys}

        # test the keys, it will set the keys to True if they are found
        _test_keys_in_dict(obj, expected_keys)

        # Check if the dictionary has all the expected keys except the ones that were not included
        for key, value in expected_keys.items():
            if key not in exclude_keys:
                assert (
                    value is True
                ), f'Key "{key}" not found in the dictionary, when using "{method}" method.'
            else:
                assert (
                    value is False
                ), f'Key "{key}" was found in the dictionary, but it should not be there, when using "{method}" method.'

        """
        to get the list of expected keys, run the following code:

            def get_keys(obj, list_keys):
                if isinstance(obj, dict):
                    for key in obj:
                        if not key in list_keys: list_keys.append(key)

                        if isinstance(obj[key], dict):
                            get_keys(obj[key], list_keys)
                        elif isinstance(obj[key], list):
                            for item in obj[key]:
                                get_keys(item, list_keys)
                elif isinstance(obj, list):
                    for item in obj:
                        get_keys(item, list_keys)

            expected_keys = []
            get_keys(qaoa.asdict(), expected_keys)
            print(expected_keys)
        """

    def test_qaoa_from_dict_and_load(self):
        """
        test loading the QAOA object from a dictionary
        methods: from_dict, load, loads
        """

        # problem
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        # run rqaoa with different devices, and save the objcets in a list
        qaoas = []
        for device in [
            create_device(location="local", name="vectorized"),
        ]:
            q = QAOA()
            q.set_device(device)
            q.set_circuit_properties(
                p=1, param_type="extended", init_type="rand", mixer_hamiltonian="x"
            )
            q.set_backend_properties(n_shots=50)
            q.set_classical_optimizer(maxiter=2, optimization_progress=True)
            q.set_exp_tags({"add_tag": "test"})
            q.set_header(
                project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
                description="test",
                run_by="oq",
                client="-",
            )

            # test that you can convert the rqaoa object to a dictionary and then load it before optimization
            _ = QAOA.from_dict(q.asdict())
            _.compile(maxcut_qubo)
            _.optimize()
            assert isinstance(
                _, QAOA
            ), "The object loaded from a dictionary is not an RQAOA object."

            # compile and optimize the original rqaoa object
            q.compile(maxcut_qubo)
            q.optimize()

            qaoas.append(q)

        # for each rqaoa object, create a new rqaoa object from dict, json string, json file,
        # and compressed json file and compare them with the original object
        for q in qaoas:
            new_q_list = []

            # get new qaoa from dict
            new_q_list.append(QAOA.from_dict(q.asdict()))
            # get new qaoa from json string
            new_q_list.append(QAOA.loads(q.dumps()))
            # get new qaoa from json file
            q.dump("test.json", prepend_id=False)
            new_q_list.append(QAOA.load("test.json"))
            os.remove("test.json")  # delete file test.json
            # get new qaoa from compressed json file
            q.dump("test.json", prepend_id=False, compresslevel=3)
            new_q_list.append(QAOA.load("test.json.gz"))
            os.remove("test.json.gz")  # delete file test.json

            for new_q in new_q_list:
                # check that the new object is an QAOA object
                assert isinstance(new_q, QAOA), "new_r is not an RQAOA object"

                # check that the attributes of the new object are of the correct type
                attributes_types = [
                    ("header", dict),
                    ("exp_tags", dict),
                    ("problem", QUBO),
                    ("result", QAOAResult),
                    ("backend_properties", BackendProperties),
                    ("classical_optimizer", ClassicalOptimizer),
                    ("circuit_properties", CircuitProperties),
                ]
                for attribute, type_ in attributes_types:
                    assert isinstance(
                        getattr(new_q, attribute), type_
                    ), f"attribute {attribute} is not type {type_}"

                # get the two objects (old and new) as dictionaries
                q_asdict = q.asdict()
                new_q_asdict = new_q.asdict()

                # compare the two dictionaries
                for key, value in q_asdict.items():
                    if key == "header":
                        assert value == new_q_asdict[key], "Header is not the same"

                    elif key == "data":
                        for key2, value2 in value.items():
                            if key2 == "input_parameters":
                                # pop key device since it is not returned completely when using asdict/dump(s)
                                value2.pop("device")
                                new_q_asdict[key][key2].pop("device")
                            if key2 == "result":
                                _compare_qaoa_results(value2, new_q_asdict[key][key2])
                            else:
                                assert (
                                    value2 == new_q_asdict[key][key2]
                                ), "{} not the same".format(key2)

                # compile and optimize the new qaoa, to check if everything is working
                new_q.compile(maxcut_qubo)
                new_q.optimize()

        # check that the RQAOA.from_dict method raises an error when using a QAOA dictionary
        error = False
        try:
            RQAOA.from_dict(q.asdict())
        except Exception:
            error = True
        assert (
            error
        ), "RQAOA.from_dict should raise an error when using a QAOA dictionary"

    def test_qaoa_evaluate_circuit(self):
        """
        test the evaluate_circuit method
        """

        # problem
        problem = MinimumVertexCover.random_instance(
            n_nodes=6, edge_probability=0.8
        ).qubo

        # run qaoa with different param_type, and save the objcets in a list
        qaoas = []
        for param_type in PARAMS_CLASSES_MAPPER.keys():
            q = QAOA()
            q.set_circuit_properties(p=3, param_type=param_type, init_type="rand")
            q.compile(problem)
            qaoas.append(q)

        # for each qaoa object, test the evaluate_circuit method
        for q in qaoas:
            # evaluate the circuit with random dict of params
            params = {
                k: np.random.rand(*v.shape)
                for k, v in q.variate_params.asdict().items()
            }
            result = q.evaluate_circuit(params)
            assert (
                abs(result["cost"]) >= 0
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` \
                should return a cost, here cost is {result['cost']}"
            assert (
                abs(result["uncertainty"]) > 0
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should return an uncertanty, \
                here uncertainty is {result['uncertainty']}"
            assert (
                len(result["measurement_results"]) > 0
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should return \
                a wavefunction when using a state-based simulator"

            # evaluate the circuit with a list of params, taking the params from the dict,
            # so we should get the same result
            params2 = []
            for value in params.values():
                params2 += value.flatten().tolist()
            result2 = q.evaluate_circuit(params2)
            assert (
                result == result2
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should return the same result \
                when passing a dict or a list of params"

            # evaluate the circuit with np.ndarray of params, taking the params from the dict,
            # so we should get the same result
            result2 = q.evaluate_circuit(np.array(params2))
            assert (
                result == result2
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should return the same result \
                when passing a dict or a list of params"

            # evaluate the circuit with the params as a QAOAVariationalBaseParams object,
            # so we should get the same result
            params_obj = deepcopy(q.variate_params)
            params_obj.update_from_raw(params2)
            result3 = q.evaluate_circuit(params_obj)
            assert (
                result == result3
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should return the same result \
                when passing a dict or a list of params"

            # run the circuit with the params manually, we should get the same result
            result4 = {}
            (
                result4["cost"],
                result4["uncertainty"],
            ) = q.backend.expectation_w_uncertainty(params_obj)
            result4["measurement_results"] = q.backend.wavefunction(params_obj)
            assert (
                result == result4
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should return the same result when \
                  passing the optimized params manually"

            # evaluate the circuit with a wrong input, it should raise an error
            error = False
            try:
                q.evaluate_circuit(1)
            except Exception:
                error = True
            assert (
                error
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should raise an error when \
                  passing a wrong input"

            # evaluate the circuit with a list longer than it should, it should raise an error
            error = False
            try:
                q.evaluate_circuit(params2 + [1])
            except Exception:
                error = True
            assert (
                error
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should raise an error when \
                passing a list longer than it should"

            # evaluate the circuit with a list shorter than it should, it should raise an error
            error = False
            try:
                q.evaluate_circuit(params2[:-1])
            except Exception:
                error = True
            assert (
                error
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should raise an error when \
                passing a list shorter than it should"

            # evaluate the circuit with a dict with a wrong key, it should raise an error
            error = False
            try:
                q.evaluate_circuit({**params, "wrong_key": 1})
            except Exception:
                error = True
            assert (
                error
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should raise an error \
                when passing a dict with a wrong key"

            # evaluate the circuit with a dict with a value longer than it should, it should raise an error
            error = False
            try:
                q.evaluate_circuit(
                    {**params, list(params.keys())[0]: np.random.rand(40)}
                )
            except Exception:
                error = True
            assert (
                error
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should raise an error when \
                passing a dict with a value longer than it should"

            # evaluate the circuit without passing any param, it should raise an error
            error = False
            try:
                q.evaluate_circuit()
            except Exception:
                error = True
            assert (
                error
            ), f"param_type={q.circuit_properties.param_type}. `evaluate_circuit` should raise an error when \
                  not passing any param"

    def test_qaoa_evaluate_circuit_shot(self):
        # problem
        problem = MinimumVertexCover.random_instance(
            n_nodes=6, edge_probability=0.8
        ).qubo

        if "qiskit.qasm_simulator" not in SUPPORTED_LOCAL_SIMULATORS:
            self.skipTest(
                reason="Qiskit QASM Simulator is not available. Please install the qiskit plugin: openqaoa-qiskit."
            )
        else:
            # check that it works with shots
            q = QAOA()
            device = create_device(location="local", name="qiskit.qasm_simulator")
            q.set_device(device)
            q.set_circuit_properties(p=3)

            # try to evaluate the circuit before compiling
            error = False
            try:
                q.evaluate_circuit()
            except Exception:
                error = True
            assert (
                error
            ), f"param_type={param_type}. `evaluate_circuit` should raise an error if the circuit is not compiled"

            # compile and evaluate the circuit, and check that the result is correct
            q.compile(problem)
            result = q.evaluate_circuit([1, 2, 1, 2, 1, 2])
            assert isinstance(
                result["measurement_results"], dict
            ), "When using a shot-based simulator, `evaluate_circuit` should return a dict of counts"
            assert (
                abs(result["cost"]) >= 0
            ), "When using a shot-based simulator, `evaluate_circuit` should return a cost"
            assert (
                abs(result["uncertainty"]) > 0
            ), "When using a shot-based simulator, `evaluate_circuit` should return an uncertainty"

            cost = cost_function(
                result["measurement_results"],
                q.backend.qaoa_descriptor.cost_hamiltonian,
                q.backend.cvar_alpha,
            )
            cost_sq = cost_function(
                result["measurement_results"],
                q.backend.qaoa_descriptor.cost_hamiltonian.hamiltonian_squared,
                q.backend.cvar_alpha,
            )
            uncertainty = np.sqrt(cost_sq - cost**2)
            assert (
                np.round(cost, 12) == result["cost"]
            ), "When using a shot-based simulator, `evaluate_circuit` not returning the correct cost"
            assert (
                np.round(uncertainty, 12) == result["uncertainty"]
            ), "When using a shot-based simulator, `evaluate_circuit` not returning the correct uncertainty"

    def test_qaoa_evaluate_circuit_analytical_sim(self):
        # problem
        problem = MinimumVertexCover.random_instance(
            n_nodes=6, edge_probability=0.8
        ).qubo

        # check that it works with analytical simulator
        q = QAOA()
        device = create_device(location="local", name="analytical_simulator")
        q.set_device(device)
        q.set_circuit_properties(p=1, param_type="standard")
        q.compile(problem)
        result = q.evaluate_circuit([1, 2])
        assert (
            abs(result["cost"]) >= 0
        ), "When using an analytical simulator, `evaluate_circuit` should return a cost"
        assert (
            result["uncertainty"] is None
        ), "When using an analytical simulator, `evaluate_circuit` should return uncertainty None"
        assert (
            result["measurement_results"] is None
        ), "When using an analytical simulator, `evaluate_circuit` should return no measurement results"

    def test_change_properties_after_compilation(self):
        device = create_device(location="local", name="vectorized")
        q = QAOA()
        q.compile(QUBO.random_instance(4))

        with self.assertRaises(ValueError):
            q.set_device(device)
        with self.assertRaises(ValueError):
            q.set_circuit_properties(
                p=1, param_type="standard", init_type="rand", mixer_hamiltonian="x"
            )
        with self.assertRaises(ValueError):
            q.set_backend_properties(prepend_state=None, append_state=None)
        with self.assertRaises(ValueError):
            q.set_classical_optimizer(
                maxiter=100, method="vgd", jac="finite_difference"
            )

    def test_numpy_serialize(self):
        np_qubo = NumberPartition([1, 2, 3]).qubo

        q = QAOA()
        q.compile(np_qubo)
        q.optimize()

        # add numpy results
        numpy_dict = {
            "000": np.int64(85),
            "100": np.int64(85),
            "010": np.int64(85),
            "111": 12,
        }
        numpy_cost = np.float64(85.123)

        q.result.optimized["intermediate"] = [
            numpy_cost,
            numpy_cost,
            numpy_cost,
            0.123,
            -123.123,
        ]
        q.result.intermediate["measurement_outcomes"] = [
            numpy_dict,
            numpy_dict,
            numpy_dict,
        ]
        q.result.optimized["measurement_outcomes"] = numpy_dict
        q.result.optimized["cost"] = numpy_cost

        q.dumps()


class TestingRQAOA(unittest.TestCase):
    """
    Unit test based testing of the RQAOA workflow class
    """

    def _test_default_values(self, x):
        """
        General function to check default values of rqaoa and qaoa
        """

        # circuit_properties
        cp = x.circuit_properties
        assert cp.param_type == "standard"
        assert cp.init_type == "ramp"
        assert cp.p == 1
        assert cp.q is None
        assert cp.mixer_hamiltonian == "x"

        # device
        d = x.device
        assert d.device_location == "local"
        assert d.device_name == "vectorized"

    def test_rqaoa_default_values(self):
        """
        Tests all default values are correct
        """
        r = RQAOA()

        assert r.rqaoa_parameters.rqaoa_type == "custom"
        assert r.rqaoa_parameters.n_cutoff == 5
        assert r.rqaoa_parameters.n_max == 1
        assert r.rqaoa_parameters.steps == 1
        assert r.rqaoa_parameters.original_hamiltonian is None
        assert r.rqaoa_parameters.counter == 0

        self._test_default_values(r)

    def test_rqaoa_compile_and_qoao_default_values(self):
        """
        Test creation of the qaoa object and its default values
        """
        r = RQAOA()
        r.compile(QUBO.random_instance(n=7))

        self._test_default_values(r._RQAOA__q)

    def test_rqaoa_from_dict_and_load(self):
        """
        test loading the RQAOA object from a dictionary
        methods: from_dict, load, loads
        """

        # problem
        maxcut_qubo = MaximumCut(
            nw.generators.fast_gnp_random_graph(n=6, p=0.6, seed=42)
        ).qubo

        # run rqaoa with different devices, and save the objcets in a list
        rqaoas = []
        for device in [
            create_device(location="local", name="vectorized"),
        ]:
            r = RQAOA()
            r.set_device(device)
            r.set_circuit_properties(
                p=1, param_type="extended", init_type="rand", mixer_hamiltonian="x"
            )
            r.set_backend_properties(n_shots=50)
            r.set_classical_optimizer(maxiter=10, optimization_progress=True)
            r.set_rqaoa_parameters(rqaoa_type="adaptive", n_cutoff=3)
            r.set_exp_tags({"tag1": "value1", "tag2": "value2"})
            r.set_header(
                project_id="8353185c-b175-4eda-9628-b4e58cb0e41b",
                description="test",
                run_by="OpenQAOA",
                provider="-",
                target="vectorized",
                cloud="local",
                client="-",
            )

            # test that you can convert the rqaoa object to a dictionary and then load it before optimization
            _ = RQAOA.from_dict(r.asdict())
            _.compile(maxcut_qubo)
            _.optimize()
            assert isinstance(
                _, RQAOA
            ), "The object loaded from a dictionary is not an RQAOA object."

            # compile and optimize the original rqaoa object
            r.compile(maxcut_qubo)
            r.optimize()

            rqaoas.append(r)

        # for each rqaoa object, create a new rqaoa object from dict, json string, json file,
        # and compressed json file and compare them with the original object
        for r in rqaoas:
            new_r_list = []

            # get new qaoa from dict
            new_r_list.append(RQAOA.from_dict(r.asdict()))
            # get new qaoa from json string
            new_r_list.append(RQAOA.loads(r.dumps()))
            # get new qaoa from json file
            r.dump("test.json", prepend_id=False)
            new_r_list.append(RQAOA.load("test.json"))
            os.remove("test.json")  # delete file test.json
            # get new qaoa from compressed json file
            r.dump("test.json", prepend_id=False, compresslevel=3)
            new_r_list.append(RQAOA.load("test.json.gz"))
            os.remove("test.json.gz")  # delete file test.json

            for new_r in new_r_list:
                # check that the new object is an RQAOA object
                assert isinstance(new_r, RQAOA), "new_r is not an RQAOA object"

                # check that the attributes of the new object are of the correct type
                attributes_types = [
                    ("header", dict),
                    ("exp_tags", dict),
                    ("problem", QUBO),
                    ("result", RQAOAResult),
                    ("backend_properties", BackendProperties),
                    ("classical_optimizer", ClassicalOptimizer),
                    ("circuit_properties", CircuitProperties),
                    ("rqaoa_parameters", RqaoaParameters),
                ]
                for attribute, type_ in attributes_types:
                    assert isinstance(
                        getattr(new_r, attribute), type_
                    ), f"attribute {attribute} is not type {type_}"

                # get the two objects (old and new) as dictionaries
                r_asdict = r.asdict()
                new_r_asdict = new_r.asdict()

                # compare the two dictionaries
                for key, value in r_asdict.items():
                    if key == "header":
                        assert value == new_r_asdict[key], "Header is not the same"

                    elif key == "data":
                        for key2, value2 in value.items():
                            if key2 == "input_parameters":
                                # pop key device
                                value2.pop("device")
                                new_r_asdict[key][key2].pop("device")
                            if key2 == "result":
                                for step in range(len(value2["intermediate_steps"])):
                                    for key3 in value2["intermediate_steps"][
                                        step
                                    ].keys():
                                        if key3 == "qaoa_results":
                                            _compare_qaoa_results(
                                                value2["intermediate_steps"][step][
                                                    "qaoa_results"
                                                ],
                                                new_r_asdict[key][key2][
                                                    "intermediate_steps"
                                                ][step]["qaoa_results"],
                                            )
                                        else:
                                            assert (
                                                value2["intermediate_steps"][step][key3]
                                                == new_r_asdict[key][key2][
                                                    "intermediate_steps"
                                                ][step][key3]
                                            ), f"{key3} is not the same"
                            else:
                                assert (
                                    value2 == new_r_asdict[key][key2]
                                ), "{} not the same".format(key2)

                # compile and optimize the new rqaoa, to check if everything is working
                new_r.compile(maxcut_qubo)
                new_r.optimize()

        # check that the Optimizer.from_dict method raises an error when using a RQAOA dictionary
        error = False
        try:
            Workflow.from_dict(r.asdict())
        except Exception:
            error = True
        assert (
            error
        ), "Optimizer.from_dict should raise an error when using a RQAOA dictionary"

    def test_change_properties_after_compilation(self):
        device = create_device(location="local", name="qiskit.shot_simulator")
        r = RQAOA()
        r.compile(QUBO.random_instance(4))

        with self.assertRaises(ValueError):
            r.set_device(device)
        with self.assertRaises(ValueError):
            r.set_circuit_properties(
                p=1, param_type="standard", init_type="rand", mixer_hamiltonian="x"
            )
        with self.assertRaises(ValueError):
            r.set_backend_properties(prepend_state=None, append_state=None)
        with self.assertRaises(ValueError):
            r.set_classical_optimizer(
                maxiter=100, method="vgd", jac="finite_difference"
            )
        with self.assertRaises(ValueError):
            r.set_rqaoa_parameters(rqaoa_type="adaptive", n_cutoff=3, n_steps=3)


if __name__ == "__main__":
    unittest.main()



"""
Energy expectation as a function of angles computed accordingly to the analytical expression for p=1.
"""
from .basebackend import QAOABaseBackend
from ..qaoa_components import (
    QAOADescriptor,
    QAOAVariationalStandardParams,
)
from ..qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from ..utilities import energy_expectation_analytical, generate_uuid, round_value


class QAOABackendAnalyticalSimulator(QAOABaseBackend):
    r"""
    A simulator class, specific for QAOA with a single layer, p=1, starting with
    a layer of Hadamards and using the X mixer. Works by calculating the expectation value
    of the given quantum circuit (specificied with beta and gamma angles) from the
    analytical formula derived in arXiv:2011.13420v2.

    Parameters
    ----------
    qaoa_descriptor: QAOADescriptor
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction and depth of the circuit.
        Note that it only works for p=1 and the X Mixer.
    """

    def __init__(
        self,
        qaoa_descriptor: QAOADescriptor,
        prepend_state=None,
        append_state=None,
        init_hadamard=True,
        cvar_alpha=1,
    ):
        # checking if not supported parameters are passed
        for k, val in {
            "Prepend_state": (prepend_state, None),
            "append_state": (append_state, None),
            "init_hadamard": (init_hadamard, True),
            "cvar_alpha": (cvar_alpha, 1),
        }.items():
            if val[0] != val[1]:
                print(
                    f"{k} is not supported for the analytical backend. {k} is set to None."
                )

        QAOABaseBackend.__init__(
            self,
            qaoa_descriptor,
            prepend_state=None,
            append_state=None,
            init_hadamard=True,
            cvar_alpha=1,
        )

        self.measurement_outcomes = (
            {}
        )  # passing empty dict for the logger since measurements are irrelevant for this backend.

        # check if conditions for the analytical formula are met
        assert self.qaoa_descriptor.p == 1, "Analytical formula only holds for p=1."

        for gatemap in self.qaoa_descriptor.mixer_qubits_singles:
            assert gatemap == "RXGateMap", "Analytical formula only holds for X mixer."

        assert (
            self.qaoa_descriptor.mixer_qubits_pairs == []
        ), "Analytical formula only holds for X mixer."

    def assign_angles(self):
        raise NotImplementedError("This method is irrelevant for this backend")

    def obtain_angles_for_pauli_list(self):
        raise NotImplementedError("This method is irrelevant for this backend")

    def qaoa_circuit(self):
        raise NotImplementedError("This method is irrelevant for this backend")

    def get_counts(self):
        raise NotImplementedError("This method is irrelevant for this backend")

    @round_value
    def expectation(self, params: QAOAVariationalBaseParams) -> float:
        """
        Compute the expectation value w.r.t the Cost Hamiltonian analytically.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.

        Returns
        -------
        float:
            Expectation value of cost operator wrt to the QAOA parameters
            according to the analytical expression for p=1.
        """
        # generate a job id
        self.job_id = generate_uuid()

        assert isinstance(
            params, QAOAVariationalStandardParams
        ), "Analytical formula only holds for standard parametrization (for now)."

        betas = params.betas
        gammas = params.gammas

        cost = energy_expectation_analytical([betas, gammas], self.cost_hamiltonian)
        return cost

    def expectation_w_uncertainty(self, params):
        raise NotImplementedError("Not implemented yet. In progress.")

    def reset_circuit(self):
        raise NotImplementedError("This method is irrelevant for this backend")

    def circuit_to_qasm(self):
        raise NotImplementedError("This method is irrelevant for this backend")



from collections import Counter
import networkx as nx

from ..utilities import check_kwargs
from .problem import Problem
from .qubo import QUBO


class MinimumVertexCover(Problem):
    """
    Creates an instance of the Minimum Vertex Cover problem.

    Parameters
    ----------
    G: nx.Graph
        The input graph as NetworkX graph instance.
    field: float
        The strength of the artificial field minimizing the size of the cover.
    penalty: float
        The strength of the penalty enforcing the cover constraint.

    Returns
    -------
    An instance of the Minimum Vertex Cover problem.
    """

    __name__ = "minimum_vertex_cover"

    def __init__(self, G, field, penalty):
        self.G = G
        self.field = field
        self.penalty = penalty

    @property
    def G(self):
        return self._G

    @G.setter
    def G(self, input_networkx_graph):
        if not isinstance(input_networkx_graph, nx.Graph):
            raise TypeError("Input problem graph must be a networkx Graph.")

        # Relabel nodes to integers starting from 0
        mapping = dict(
            zip(input_networkx_graph, range(input_networkx_graph.number_of_nodes()))
        )
        self._G = nx.relabel_nodes(input_networkx_graph, mapping)

    @property
    def field(self):
        return self._field

    @field.setter
    def field(self, input_field):
        if not isinstance(input_field, int) and not isinstance(input_field, float):
            raise TypeError(
                "The input parameter, field, has to be of type float or int"
            )

        self._field = input_field

    @property
    def penalty(self):
        return self._penalty

    @penalty.setter
    def penalty(self, input_penalty):
        if not isinstance(input_penalty, int) and not isinstance(input_penalty, float):
            raise TypeError(
                "The input parameter, penalty, has to be of type float or int"
            )

        self._penalty = input_penalty

    @staticmethod
    def random_instance(**kwargs):
        """
        Creates a random instance of the Minimum Vertex Cover problem, whose graph is
        random following the Erdos-Renyi model. By default the artificial field is
        set to 1.0 and the default penalty os taken to be 10 times larger.

        Parameters
        ----------
        **kwargs:
            Required keyword arguments are:

            n_nodes: int
                The number of nodes (vertices) in the graph.
            edge_probability: float
                The probability with which an edge is added to the graph.

        Returns
        -------
        A random instance of the Minimum Vertex Cover problem.
        """

        n_nodes, edge_probability = check_kwargs(
            ["n_nodes", "edge_probability"], [None, None], **kwargs
        )
        seed = kwargs.get("seed", None)
        G = nx.generators.random_graphs.fast_gnp_random_graph(
            n=n_nodes, p=edge_probability, seed=seed
        )

        DEFAULT_FIELD = 1.0
        DEFAULT_PENALTY = 10

        return MinimumVertexCover(G, DEFAULT_FIELD, DEFAULT_PENALTY)

    def terms_and_weights(self):
        """
        Creates the terms and weights for the Minimum Vertex Cover problem

        Returns
        -------
        terms_weights: tuple(list[list],list[float])
            Tuple containing list of terms and list of weights.
        """

        # Number of nodes
        n_nodes = self.G.number_of_nodes()

        # Number of edges
        edges = list(self.G.edges())

        # Connectivity of each node
        node_repetition = [e for edge in edges for e in edge]
        connectivity = dict(Counter(node_repetition))

        # Quadratic interation from penalty term
        quadratic_interaction = [(list(e), self.penalty / 4) for e in edges]

        # Linear terms from the artificial field
        linear_interaction = [
            ([i], -self.field / 2 + connectivity[i] * self.penalty / 4)
            if connectivity.get(i) is not None
            else ([i], -self.field / 2)
            for i in range(n_nodes)
        ]

        # Constant term
        constant_term = [([], n_nodes * self.field / 2 + len(edges) * self.penalty / 4)]

        # Generate tuple containing a list with the terms and a list with the weights
        terms_weights = tuple(
            zip(*(quadratic_interaction + linear_interaction + constant_term))
        )

        # Unzip to retrieve terms and weights in separate sequences
        return terms_weights

    @property
    def qubo(self):
        """
        Returns the QUBO encoding of this problem.

        Returns
        -------
        The QUBO encoding of this problem.
        """

        # Extract terms and weights from the problem definition
        terms, weights = self.terms_and_weights()

        return QUBO(
            self.G.number_of_nodes(), list(terms), list(weights), self.problem_instance
        )



import sys
from importlib.metadata import entry_points


def plugin_finder_dict() -> list:
    """
    This function searches the openqaoa.plugins entry point group to see if
    any of the openqaoa plugins have been installed. If they are, this function
    returns a dictionary whose keys are the names of the plugin libraries and
    whose values are the plugin's utilities module.
    """

    if sys.version_info >= (3, 10):
        available_plugins = entry_points().select(group="openqaoa.plugins")
    else:
        available_plugins = entry_points().get("openqaoa.plugins", [])

    output_dict = dict()
    for each_plugin_entry_point in available_plugins:
        try:
            output_dict[each_plugin_entry_point.name] = each_plugin_entry_point.load()
        except ModuleNotFoundError as e:
            print(
                "The {} module has not been installed.".format(
                    each_plugin_entry_point.name
                )
            )
            print(e)
        except AttributeError:
            print(
                "An error has occured when trying to attach the {} plugin.".format(
                    each_plugin_entry_point.name
                )
            )

    return output_dict


#     output_dict = dict()

#     try:
#         available_plugins = entry_points()['openqaoa.plugins']

#         for each_plugin_entry_point in available_plugins:
#             try:
#                 output_dict[each_plugin_entry_point.name] = each_plugin_entry_point.load()
#                 print(output_dict)
#             except ModuleNotFoundError:
#                 print("The {} module has not been installed.".format(each_plugin_entry_point.name))
#             except AttributeError:
#                 print("An error has occured when trying to attach the {} plugin.".format(each_plugin_entry_point.name))

#     except KeyError:
#         print("No plugins were found.")

#     return output_dict

PLUGIN_DICT = plugin_finder_dict()



from __future__ import annotations
from copy import deepcopy
import numpy as np

from ..qaoa_components import QAOAVariationalExtendedParams
from openqaoa.qaoa_components.variational_parameters.variational_baseparams import (
    QAOAVariationalBaseParams,
)
from ..backends.basebackend import QAOABaseBackend
from ..optimizers.logger_vqa import Logger
from ..backends.cost_function import cost_function


def update_and_compute_expectation(
    backend_obj: QAOABaseBackend, params: QAOAVariationalBaseParams, logger: Logger
):
    """
    Helper function that returns a callable that takes in a list/nparray of raw parameters.
    This function will handle:

        #. Updating logger object with `logger.log_variables`
        #. Updating variational parameters with `update_from_raw`
        #. Computing expectation with `backend_obj.expectation`

    Parameters
    ----------
    backend_obj: QAOABaseBackend
        `QAOABaseBackend` object that contains information about the
        backend that is being used to perform the QAOA circuit
    params : QAOAVariationalBaseParams
        `QAOAVariationalBaseParams` object containing variational angles.
    logger: Logger
        Logger Class required to log information from the evaluations
        required for the jacobian/hessian computation.
    Returns
    -------
    out:
        A callable that accepts a list/array of parameters, and returns the
        computed expectation value.
    """

    def fun(args, n_shots=None):
        current_total_eval = logger.func_evals.best[0]
        current_total_eval += 1
        current_jac_eval = logger.jac_func_evals.best[0]
        current_jac_eval += 1
        logger.log_variables(
            {"func_evals": current_total_eval, "jac_func_evals": current_jac_eval}
        )
        params.update_from_raw(args)

        n_shots_dict = {"n_shots": n_shots} if n_shots else {}
        return backend_obj.expectation(params, **n_shots_dict)

    return fun


def update_and_get_counts(
    backend_obj: QAOABaseBackend, params: QAOAVariationalBaseParams, logger: Logger
):
    """
    Helper function that returns a callable that takes in a list/nparray of
    raw parameters.
    This function will handle:

        #. Updating logger object with `logger.log_variables`
        #. Updating variational parameters with `update_from_raw`
        #. Getting the counts dictonary with `backend_obj.get_counts`

    PARAMETERS
    ----------
    backend_obj: QAOABaseBackend
        `QAOABaseBackend` object that contains information about the backend
        that is being used to perform the QAOA circuit
    params : QAOAVariationalBaseParams
        `QAOAVariationalBaseParams` object containing variational angles.
    logger: Logger
        Logger Class required to log information from the evaluations
        required for the jacobian/hessian computation.

    Returns
    -------
    out:
        A callable that accepts a list/array of parameters,
        and returns the counts dictonary.
    """

    def fun(args, n_shots=None):
        current_total_eval = logger.func_evals.best[0]
        current_total_eval += 1
        current_jac_eval = logger.jac_func_evals.best[0]
        current_jac_eval += 1
        logger.log_variables(
            {"func_evals": current_total_eval, "jac_func_evals": current_jac_eval}
        )
        params.update_from_raw(args)

        n_shots_dict = {"n_shots": n_shots} if n_shots else {}
        return backend_obj.get_counts(params, **n_shots_dict)

    return fun


def derivative(
    backend_obj: QAOABaseBackend,
    params: QAOAVariationalBaseParams,
    logger: Logger,
    derivative_type: str = None,
    derivative_method: str = None,
    derivative_options: dict = None,
):
    """
    Returns a callable function that calculates the gradient according
    to the specified `gradient_method`.

    Parameters
    ----------
    backend_obj: QAOABaseBackend
        `QAOABaseBackend` object that contains information about the
        backend that is being used to perform the QAOA circuit
    params : QAOAVariationalBaseParams
        `QAOAVariationalBaseParams` object containing variational angles.
    logger: Logger
        Logger Class required to log information from the evaluations
        required for the jacobian/hessian computation.
    derivative_type : str
        Type of derivative to compute. Either `gradient` or `hessian`.
    derivative_method : str
        Computational method of the derivative.
        Either `finite_difference`, `param_shift`, `stoch_param_shift`, or `grad_spsa`.
    derivative_options : dict
        Dictionary containing options specific to each `derivative_method`.
    cost_std :
        object that computes expectation values when executed. Standard parametrisation.
    cost_ext :
        object that computes expectation values when executed. Extended parametrisation.
        Mainly used to compute parameter shifts at each individual gate,
        which is summed to recover the parameter shift for a parametrised layer.
    Returns
    -------
    out:
        The callable derivative function of the cost function,
        generated based on the `derivative_type`, `derivative_method`,
        and `derivative_options` specified.
    """
    # Default derivative_options used if none are specified.
    default_derivative_options = {
        "stepsize": 0.00001,
        "n_beta_single": -1,
        "n_beta_pair": -1,
        "n_gamma_single": -1,
        "n_gamma_pair": -1,
    }

    derivative_options = (
        {**default_derivative_options, **derivative_options}
        if derivative_options is not None
        else default_derivative_options
    )

    # cost_std = derivative_dict['cost_std']
    # cost_ext = derivative_dict['cost_ext']
    params_ext = QAOAVariationalExtendedParams.empty(backend_obj.qaoa_descriptor)

    derivative_types = ["gradient", "gradient_w_variance", "hessian"]
    assert (
        derivative_type in derivative_types
    ), "Unknown derivative type specified - please choose between " + str(
        derivative_types
    )

    derivative_methods = [
        "finite_difference",
        "param_shift",
        "stoch_param_shift",
        "grad_spsa",
    ]
    assert derivative_method in derivative_methods, (
        "Unknown derivative computation method specified - please choose between "
        + str(derivative_methods)
    )

    params = deepcopy(params)

    if derivative_type == "gradient":
        if derivative_method == "finite_difference":
            out = grad_fd(backend_obj, params, derivative_options, logger)
        elif derivative_method == "param_shift":
            assert (
                params.__class__.__name__ == "QAOAVariationalStandardParams"
            ), f"{params.__class__.__name__} not supported - only Standard Parametrisation is supported for parameter shift/stochastic parameter shift for now."
            out = grad_ps(backend_obj, params, params_ext, derivative_options, logger)
        elif derivative_method == "stoch_param_shift":
            assert (
                params.__class__.__name__ == "QAOAVariationalStandardParams"
            ), f"{params.__class__.__name__} not supported - only Standard Parametrisation is supported for parameter shift/stochastic parameter shift for now."
            out = grad_ps(
                backend_obj,
                params,
                params_ext,
                derivative_options,
                logger,
                stochastic=True,
            )
        elif derivative_method == "grad_spsa":
            out = grad_spsa(backend_obj, params, derivative_options, logger)

    elif derivative_type == "gradient_w_variance":
        if derivative_method == "finite_difference":
            out = grad_fd(
                backend_obj, params, derivative_options, logger, variance=True
            )
        elif derivative_method == "param_shift":
            assert (
                params.__class__.__name__ == "QAOAVariationalStandardParams"
            ), f"{params.__class__.__name__} not supported - only Standard Parametrisation is supported for parameter shift/stochastic parameter shift for now."
            out = grad_ps(
                backend_obj,
                params,
                params_ext,
                derivative_options,
                logger,
                variance=True,
            )
        elif derivative_method == "stoch_param_shift":
            assert (
                params.__class__.__name__ == "QAOAVariationalStandardParams"
            ), f"{params.__class__.__name__} not supported - only Standard Parametrisation is supported for parameter shift/stochastic parameter shift for now."
            out = grad_ps(
                backend_obj,
                params,
                params_ext,
                derivative_options,
                logger,
                stochastic=True,
                variance=True,
            )
        elif derivative_method == "grad_spsa":
            out = grad_spsa(
                backend_obj, params, derivative_options, logger, variance=True
            )

    elif derivative_type == "hessian":
        if derivative_method == "finite_difference":
            out = hessian_fd(backend_obj, params, derivative_options, logger)
        else:
            raise ValueError(
                "Only support hessian derivative method is finite_difference. Your choice: {}".format(
                    derivative_method
                )
            )

    return out


def __create_n_shots_list(n_params, n_shots):
    """
    Creates a list of `n_shots` for each parameter in the variational circuit.

    Parameters
    ----------
    n_params : int
        Number of parameters in the variational circuit.
    n_shots : int or list
        Number of shots to use for each parameter in the variational circuit.
        If `n_shots` is an integer, then the same number of shots is used for each parameter.
        If `n_shots` is a list, then the length of the list must be equal to `n_params`
        and we return the list.

    Returns
    -------
    out:
        List of `n_shots` for each parameter in the variational circuit.
    """

    # If n_shots is a list, then it must be of length n_params,
    # else create a list of length n_params with all elements equal to n_shots
    if isinstance(n_shots, list):
        assert (
            len(n_shots) == n_params
        ), "n_shots must be a list of length equal to the number of parameters."
        n_shots_list = n_shots
    elif isinstance(n_shots, int) or n_shots is None:
        n_shots_list = [n_shots] * n_params
    else:
        raise ValueError("n_shots must be either an integer or a list of integers.")

    return n_shots_list


def __create_n_shots_ext_list(n_params, n_associated_params, n_shots):
    """
    Creates a list of number of shots for each parameter in the extended parametrisation.
    If n_shots is a integer, then it is used for all extended parameters. So, we create a list of length sum(n_associated_params)
    with all elements equal to n_shots. (sum(n_associated_params) = number of extended params)
    If n_shots is a list, then this list tell us the number of shots for each standard parameter. We convert this list to a list of
    number of shots for each extended parameter. Each standard parameter has a different number of associated extended parameters,
    `n_associated_params` helps us with this. Each element of `n_associated_params` is the number of associated extended parameters to each coefficient.
    And we know that each standard parameter has 2 associated coefficients (mixer_1q, mixer_2q, cost_1q, cost_2q).

    Parameters
    ----------
    n_associated_params : list
        List of integers, where each integer is the number of associated parameters in the extended parametrisation for each coefficient.
        The sum of all elements in this list is equal to the number of parameters in the extended parametrisation.
    n_params : int
        Number of parameters in the standard parametrisation.
    n_shots : int or list
        Number of shots to use for each parameter in the standard parametrisation.
        If an integer, then it is used for all the extended parameters.
        If a list, then it must be of length sum(n_associated_params).

    Returns
    -------
    n_shots_ext_list : list
        List of integers, where each integer is the number of shots to use for each parameter in the extended parametrisation.
    """

    # If n_shots is a list, then it must be of length n_params, else create a list of length n_params with all elements equal to n_shots
    if isinstance(n_shots, list):
        assert (
            len(n_shots) == n_params
        ), "n_shots must be a list of length equal to the number of parameters."

        # transform n_shots list (which has length n_params) into a list of the same length of n_associated_params
        n_shots = np.array(n_shots)
        n_shots = (
            n_shots.reshape(2, len(n_shots) // 2)
            .repeat(2, axis=0)
            .reshape(n_shots.size * 2)
        )  # repeat each element twice in the following way: if we have [1,2,3,4] we get [1,2,1,2,3,4,3,4]

        # create the list of n_shots for each parameter in the extended parametrisation.
        # For each parameter of each coefficient we have a number of shots (`n_shots` list),
        # we need to repeat this number of shots the number of associated extended parameters
        # that each coefficient has (`n_associated_params` list).
        n_shots_list = [
            shots for r, shots in zip(n_associated_params, n_shots) for _ in range(r)
        ]
    elif isinstance(n_shots, int) or n_shots is None:
        n_shots_list = [n_shots] * np.sum(n_associated_params)
    else:
        raise ValueError("n_shots must be either an integer or a list of integers.")

    return n_shots_list


def __gradient(args, backend_obj, params, logger, variance):
    """
    Returns a callable function that computes the gradient as `constant*(fun(args + vect_eta) - fun(args - vect_eta))`
    and its variance (if variance=True).
    Where fun is the function that we want to differentiate,
    vect_eta is a vector of length n_params, and constant is a
    constant that depends on the derivative method.

    Parameters
    ----------
    args : np.array
        List of arguments to pass to the function that we want to differentiate.
    backend_obj : QAOABaseBackend
        `QAOABaseBackend` object that contains information about the backend that
        is being used to perform the QAOA circuit
    params : QAOAVariationalBaseParams
        `QAOAVariationalBaseParams` object containing variational angles.
    logger: Logger
        Logger Class required to log information from the evaluations
        required for the jacobian/hessian computation.
    variance : bool
        If True, then the variance of the gradient is also computed.

    Returns
    -------
    out:
        Callable function that computes the gradient and its variance (if variance=True).
    """

    def fun_w_variance(vect_eta, constant, n_shots=None):
        """
        Computes the gradient and its variance as `constant*(fun(args + vect_eta) - fun(args - vect_eta))`.

        Parameters
        ----------
        vect_eta : np.array
            Vector of length n_params.
        constant : float
            Constant that depends on the derivative method.
        n_shots : int
            Number of shots to use when calling fun().

        Returns
        -------
        out:
            Gradient and its variance.
        """

        # get value of eta, the function to get counts, hamiltonian, and alpha
        fun = update_and_get_counts(backend_obj, params, logger)
        hamiltonian = backend_obj.qaoa_descriptor.cost_hamiltonian
        alpha = backend_obj.cvar_alpha

        # get counts f(x+eta/2) and f(x-eta/2)
        counts_i_dict = fun(args - vect_eta, n_shots=n_shots)
        counts_f_dict = fun(args + vect_eta, n_shots=n_shots)

        # compute cost for each state in the counts dictionaries
        costs_dict = {
            key: cost_function({key: 1}, hamiltonian, alpha)
            for key in counts_i_dict.keys() | counts_f_dict.keys()
        }

        # for each count get the cost and create a list of shot costs
        eval_i_list = [
            costs_dict[key]
            for key, value in counts_i_dict.items()
            for _ in range(value)
        ]
        eval_f_list = [
            costs_dict[key]
            for key, value in counts_f_dict.items()
            for _ in range(value)
        ]

        # check if the number of shots used in the simulator / QPU is equal to n_shots
        assert (
            len(eval_i_list) == n_shots and len(eval_f_list) == n_shots
        ), "This backend does not support changing the number of shots."

        # compute a list of gradients of one shot cost
        grad_list = np.real(constant * (np.array(eval_f_list) - np.array(eval_i_list)))

        # return average and variance for the gradient for this argument
        return np.mean(grad_list), np.var(grad_list)

    def fun(vect_eta, constant, n_shots=None):
        """
        Computes the gradient as `constant*(fun(args + vect_eta) - fun(args - vect_eta))`.

        Parameters
        ----------
        vect_eta : np.array
            Vector of length n_params.
        constant : float
            Constant that depends on the derivative method.
        n_shots : int
            Number of shots to use when calling fun().

        Returns
        -------
        out:
            Gradient.
        """
        fun = update_and_compute_expectation(backend_obj, params, logger)
        return (
            constant
            * (
                fun(args + vect_eta, n_shots=n_shots)
                - fun(args - vect_eta, n_shots=n_shots)
            ),
            0,
        )

    if variance:
        return fun_w_variance
    else:
        return fun


def grad_fd(backend_obj, params, gradient_options, logger, variance: bool = False):
    """
    Returns a callable function that calculates the gradient (and its variance if `variance=True`)
    with the finite difference method.

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed.
    params : `QAOAVariationalBaseParams`
        parameters of the variational circuit.
    gradient_options : `dict`
        stepsize :
            Stepsize of finite difference.
    logger : `Logger`
        logger object to log the number of function evaluations.
    variance : `bool`
        If True, the variance of the gradient is also computed.
        If False, only the gradient is computed.

    RETURNS
    -------
    grad_fd_func: `Callable`
        Callable derivative function.
    """

    # Set default value of eta
    eta = gradient_options["stepsize"]

    def grad_fd_func(args, n_shots=None):
        # get the function to compute the gradient and its variance
        __gradient_function = __gradient(args, backend_obj, params, logger, variance)

        # if n_shots is int or None create a list with len of args
        # (if it is none, it will use the default n_shots)
        n_shots_list = __create_n_shots_list(len(args), n_shots)

        # lists of gradients and variances for each argument, initialized with zeros
        grad, var = np.zeros(len(args)), np.zeros(len(args))

        for i in range(len(args)):
            # vector and constant to compute the gradient for the i-th argument
            vect_eta = np.zeros(len(args))
            vect_eta[i] = eta / 2
            const = 1 / eta

            # Finite diff. calculation of gradient
            grad[i], var[i] = __gradient_function(
                vect_eta, const, n_shots_list[i]
            )  # const*[f(args + vect_eta) - f(args - vect_eta)]

        # if variance is True, add the number of shots per argument to the logger
        if variance:
            logger.log_variables({"n_shots": n_shots_list})

        if variance:
            return (
                grad,
                var,
                2 * sum(n_shots_list),
            )  # return gradient, variance, and total number of shots
        else:
            return grad  # return gradient

    return grad_fd_func


def grad_ps(
    backend_obj,
    params,
    params_ext,
    gradient_options,
    logger,
    stochastic: bool = False,
    variance: bool = False,
):
    """
    If `stochastic=False` returns a callable function that calculates the gradient
    (and its variance if `variance=True`) with the parameter shift method.
    If `stochastic=True` returns a callable function that approximates the gradient
    (and its variance if `variance=True`) with the stochastic parameter shift method,
    which samples (n_beta_single, n_beta_pair, n_gamma_single, n_gamma_pair) gates at
    each layer instead of all gates. See "Algorithm 4" of https://arxiv.org/pdf/1910.01155.pdf.
    By convention, (n_beta_single, n_beta_pair, n_gamma_single, n_gamma_pair) = (-1, -1, -1, -1)
    will sample all gates (which is then equivalent to the full parameter shift rule).

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed.
    params : `QAOAVariationalStandardParams`
        variational parameters object, standard parametrisation.
    params_ext : `QAOAVariationalExtendedParams`
        variational parameters object, extended parametrisation.
    gradient_options :
        * 'n_beta_single': Number of single-qubit mixer gates to sample for the stochastic parameter shift.
        * 'n_beta_pair': Number of two-qubit mixer gates to sample for the stochastic parameter shift.
        * 'n_gamma_single': Number of single-qubit cost gates to sample for the stochastic parameter shift.
        * 'n_gamma_pair': Number of two-qubit cost gates to sample for the stochastic parameter shift.
    logger : `Logger`
        logger object to log the number of function evaluations.
    variance : `bool`
        If True, the variance of the gradient is also computed.
        If False, only the gradient is computed.

    RETURNS
    -------
    grad_ps_func:
        Callable derivative function.
    """
    # TODO : handle Fourier parametrisation

    # list of list of the coefficients (mixer_1q, mixer_2q, cost_1q, cost_2q)
    params_coeffs = [
        params.mixer_1q_coeffs,
        params.mixer_2q_coeffs,
        params.cost_1q_coeffs,
        params.cost_2q_coeffs,
    ]
    # list of coefficients for all the extended parameters
    coeffs_list = [
        x for coeffs in params_coeffs for x in coeffs for _ in range(params.p)
    ]

    # create a list of how many extended parameters are associated with each coefficient
    n_associated_params = np.repeat([len(x) for x in params_coeffs], params.p)
    # with the `n_associated_params` list, add a 0 in the first position and
    # sum the list cumulatively (so that this list indicate which gate is associated with which coefficient)
    l = np.insert(
        np.cumsum(n_associated_params), 0, 0
    )  # the i-th coefficient is associated with extended parameters with indices in range [l[i], l[i+1]]

    def grad_ps_func(args, n_shots=None):
        # Convert standard to extended parameters before applying parameter shift
        args_ext = params.convert_to_ext(args)

        # get the function to compute the gradient and its variance
        __gradient_function = __gradient(
            args_ext, backend_obj, params_ext, logger, variance
        )

        # we call the function that returns the number of shots for each extended parameter,
        # giving the number of shots for each standard parameter (n_shots)
        n_shots_list = __create_n_shots_ext_list(
            len(args), n_associated_params, n_shots
        )

        # compute the gradient (and variance) of all extended parameters
        # with the stochastic parameter shift method lists of gradients and variances
        # for each argument (extended), initialized with zeros
        grad_ext, var_ext = np.zeros(len(args_ext)), np.zeros(len(args_ext))
        # variable to count the number of shots used
        n_shots_used = 0
        # Apply parameter shifts
        for i, sample_param_bool in enumerate(__sample_params()):
            # __sample_params is only relevant if stochastic=True, if stochastic=False
            # we loop over all extended parameters. __sample_params() returns a list with True or False
            # for each extended parameter (`i` is the index of the extended parameter,
            # and `sample_param_bool` is True if the parameter is sampled)
            if not sample_param_bool:
                # (only relevant if stochastic=True) if the parameter is not
                # sampled (`sample_param_bool==False`), skip it
                continue
            r = coeffs_list[i]
            vect_eta = np.zeros(len(args_ext))
            vect_eta[i] = np.pi / (4 * r)
            grad_ext[i], var_ext[i] = __gradient_function(
                vect_eta, r, n_shots_list[i]
            )  # r*[f(args + vect_eta) - f(args - vect_eta)]
            n_shots_used += (
                n_shots_list[i] * 2 if n_shots_list[i] != None else 0
            )  # we multiply by 2 because we call the function twice to compute the gradient

        ## convert extended form back into standard form
        # sum all the gradients for each parameter of each coefficient according to the indices in l,
        # and rehape the array to 4 rows and p columns, first row is mixer_1q, second is mixer_2q,
        # third is cost_1q, fourth is cost_2q
        grad = np.array(
            [np.sum(grad_ext[l[i] : l[i + 1]]) for i in range(len(l) - 1)]
        ).reshape(4, params.p)
        # summing 1q with 2q gradients (first row with second row, third row with fourth row),
        # to get the gradient for each parameter in the standard form
        grad = np.concatenate(
            (grad[0] + grad[1], grad[2] + grad[3])
        )  # now we have the gradient in the standard form (one gradient for each standard parameter)
        # repeat the same for the variances
        var = np.array(
            [np.sum(var_ext[l[i] : l[i + 1]]) for i in range(len(l) - 1)]
        ).reshape(4, params.p)
        var = np.concatenate((var[0] + var[1], var[2] + var[3]))

        # if variance is True, add the number of shots per argument (standard) to the logger
        if variance:
            logger.log_variables({"n_shots": __create_n_shots_list(len(args), n_shots)})

        if variance:
            return (
                grad,
                var,
                n_shots_used,
            )  # return gradient, variance, and total number of shots
        else:
            return grad  # return gradient

    def __sample_params():
        """
        Generator that returns a list with True or False for each extended parameter,
        if False the parameter is not sampled, meaning that the gradient is not computed for that parameter.
        For parameter shift method (not stochastic), all parameters are sampled.
        For stochastic parameter shift method, the parameters are sampled according
        to the number given by the user for each parameter.
        """

        if stochastic:
            # if stochastic is True, we sample the parameters according to the number given by the
            # user for each parameter, this information is in the `gradient_options` dictionary.
            # list of the names of the parameters and list of the number of gates for each parameter
            names_params = [
                "n_beta_single",
                "n_beta_pair",
                "n_gamma_single",
                "n_gamma_pair",
            ]
            n_sample = [gradient_options[x] for x in names_params]

            # check if the number of gates to sample is valid and if it is -1
            # then set it to the number of gates in the full parameter shift rule
            for i, x in enumerate(n_sample):
                y = len(params_coeffs[i])
                assert (
                    -1 <= x <= y
                ), f"Invalid jac_options['{names_params[i]}'], it must be between -1 and {y}, but {x} is passed."
                if x == -1:
                    n_sample[i] = y

            # create a list of how many gates are associated with each coefficient
            n_sample = np.repeat(n_sample, params.p)

            # sample the parameters according to the number of gates to sample (this information is in n_sample)
            sampled_list = np.array([])
            for i, n in enumerate(n_sample):
                # randomly choose n parameters to sample in the i-th group of standard parameters
                sampled_indices = np.random.choice(range(l[i], l[i + 1]), n, False)
                # create a list of booleans, where True means that the parameter is sampled
                sampled_list = np.append(
                    sampled_list,
                    np.array([i in sampled_indices for i in range(l[i], l[i + 1])]),
                )
        else:
            # if stochastic is False, sample all the parameters (there is one parameter for each coefficient)
            sampled_list = np.ones(len(coeffs_list), dtype=bool)

        return sampled_list

    return grad_ps_func


def grad_spsa(backend_obj, params, gradient_options, logger, variance: bool = False):
    """
    Returns a callable function that calculates the gradient approxmiation with the
    Simultaneous Perturbation Stochastic Approximation (SPSA) method.

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed.
    params : `QAOAVariationalBaseParams`
        variational parameters object.
    gradient_options : `dict`
        gradient_stepsize :
            stepsize of stochastic shift.
    logger : `Logger`
        logger object to log the number of function evaluations.

    RETURNS
    -------
    grad_spsa_func: `Callable`
        Callable derivative function.

    """
    eta = gradient_options["stepsize"]

    def grad_spsa_func(args, n_shots=None):
        # if variance is True, add the number of shots per argument to the logger
        if variance:
            logger.log_variables({"n_shots": [n_shots]})

        # get the function to compute the gradient and its variance
        __gradient_function = __gradient(args, backend_obj, params, logger, variance)

        # vector and constant to compute the gradient and its variance
        delta = 2 * np.random.randint(0, 2, size=len(args)) - 1
        vector_eta = delta * eta / 2
        const = 1 / eta

        # compute the gradient and its variance: const*[f(args + vect_eta) - f(args - vect_eta)]
        grad, var = __gradient_function(vector_eta, const, n_shots)

        if variance:
            return (
                grad * delta,
                var * np.abs(delta),
                2 * n_shots,
            )  # return the gradient, its variance and the total number of shots
        else:
            return grad * delta  # return the gradient

    return grad_spsa_func


def hessian_fd(backend_obj, params, hessian_options, logger):
    """
    Returns a callable function that calculates the hessian with the finite difference method.

    PARAMETERS
    ----------
    backend_obj : `QAOABaseBackend`
        backend object that computes expectation values when executed.
    params : `QAOAVariationalBaseParams`
        variational parameters object.
    hessian_options :
        hessian_stepsize :
            stepsize of finite difference.
    logger : `Logger`
        logger object to log the number of function evaluations.

    RETURNS
    -------
    hessian_fd_func:
        Callable derivative function.

    """

    eta = hessian_options["stepsize"]
    fun = update_and_compute_expectation(backend_obj, params, logger)

    def hessian_fd_func(args):
        hess = np.zeros((len(args), len(args)))

        for i in range(len(args)):
            for j in range(len(args)):
                vect_eta1 = np.zeros(len(args))
                vect_eta2 = np.zeros(len(args))
                vect_eta1[i] = 1
                vect_eta2[j] = 1

                if i == j:
                    # Central diff. hessian diagonals (https://v8doc.sas.com/sashtml/ormp/chap5/sect28.htm)
                    hess[i][i] = (
                        -fun(args + 2 * eta * vect_eta1)
                        + 16 * fun(args + eta * vect_eta1)
                        - 30 * fun(args)
                        + 16 * fun(args - eta * vect_eta1)
                        - fun(args - 2 * eta * vect_eta1)
                    ) / (12 * eta**2)
                # grad_diff[i] = (grad_fd_ext(params + (eta/2)*vect_eta)[i] - grad_fd_ext(params - (eta/2)*vect_eta)[i])/eta
                else:
                    hess[i][j] = (
                        fun(args + eta * vect_eta1 + eta * vect_eta2)
                        - fun(args + eta * vect_eta1 - eta * vect_eta2)
                        - fun(args - eta * vect_eta1 + eta * vect_eta2)
                        + fun(args - eta * vect_eta1 - eta * vect_eta2)
                    ) / (4 * eta**2)

        return hess

    return hessian_fd_func



import unittest
from random import randint, random
import numpy as np
from openqaoa.problems import (
    NumberPartition,
    QUBO,
    TSP,
    TSP_LP,
    Knapsack,
    ShortestPath,
    SlackFreeKnapsack,
    MaximumCut,
    MinimumVertexCover,
    VRP,
    PortfolioOptimization,
    MIS,
    BinPacking,
    SK,
    KColor,
)
from openqaoa.utilities import convert2serialize
from openqaoa.problems.helper_functions import create_problem_from_dict


class TestQUBO(unittest.TestCase):
    """Tests for the QUBO class"""

    def __generate_random_problems(self):
        seed = 0
        rng = np.random.default_rng(seed)
        problems_random_instances = {
            "tsp": TSP.random_instance(n_cities=int(rng.integers(2, 10)), seed=seed),
            "tsp_lp": TSP_LP.random_instance(
                n_nodes=int(rng.integers(2, 10)), seed=seed
            ),
            "number_partition": NumberPartition.random_instance(
                n_numbers=int(rng.integers(2, 10)), seed=seed
            ),
            "maximum_cut": MaximumCut.random_instance(
                n_nodes=int(rng.integers(2, 10)),
                edge_probability=rng.random(),
                seed=seed,
            ),
            "knapsack": Knapsack.random_instance(
                n_items=int(rng.integers(2, 10)), seed=seed
            ),
            "slack_free_knapsack": SlackFreeKnapsack.random_instance(
                n_items=int(rng.integers(2, 10)), seed=seed
            ),
            "minimum_vertex_cover": MinimumVertexCover.random_instance(
                n_nodes=int(rng.integers(2, 10)),
                edge_probability=rng.random(),
                seed=seed,
            ),
            "shortest_path": ShortestPath.random_instance(
                n_nodes=int(rng.integers(3, 10)),
                edge_probability=rng.random(),
                seed=seed,
            ),
            "vehicle_routing": VRP.random_instance(seed=seed),
            "maximal_independent_set": MIS.random_instance(
                n_nodes=int(rng.integers(3, 10)),
                edge_probability=rng.random(),
                seed=seed,
            ),
            "bin_packing": BinPacking.random_instance(seed=seed),
            "portfolio_optimization": PortfolioOptimization.random_instance(seed=seed),
            "sherrington_kirkpatrick": SK.random_instance(
                n_nodes=int(rng.integers(2, 10)), seed=seed
            ),
            "k_color": KColor.random_instance(
                n_nodes=int(rng.integers(3, 8)), k=int(rng.integers(2, 5)), seed=seed
            ),
        }
        qubo_random_instances = {
            k: v.qubo for k, v in problems_random_instances.items()
        }
        qubo_random_instances["generic_qubo"] = QUBO.random_instance(
            int(rng.integers(2, 10)), seed=seed
        )
        return problems_random_instances, qubo_random_instances

    def test_problem_instance(self):
        """
        Test problem instance method of the QUBO class.
        From the random instance of all the different problems, we generate the
        QUBO problem out of it and then we check if the problem instance
        attribute is correct, by comparing the keys of the problem instance
        with the expected keys.
        """

        _, qubos = self.__generate_random_problems()

        expected_keys = {
            "tsp": ["problem_type", "n_cities", "G", "A", "B"],
            "tsp_lp": [
                "problem_type",
                "G",
                "n_vehicles",
                "pos",
                "depot",
                "subtours",
                "method",
                "penalty",
            ],
            "number_partition": ["problem_type", "numbers", "n_numbers"],
            "maximum_cut": ["problem_type", "G"],
            "knapsack": [
                "problem_type",
                "values",
                "weights",
                "weight_capacity",
                "penalty",
                "n_items",
            ],
            "slack_free_knapsack": [
                "problem_type",
                "values",
                "weights",
                "weight_capacity",
                "penalty",
                "n_items",
            ],
            "minimum_vertex_cover": ["problem_type", "G", "field", "penalty"],
            "shortest_path": ["problem_type", "G", "source", "dest"],
            "vehicle_routing": [
                "problem_type",
                "G",
                "n_vehicles",
                "pos",
                "depot",
                "subtours",
                "method",
                "penalty",
            ],
            "maximal_independent_set": ["problem_type", "G", "penalty"],
            "bin_packing": [
                "problem_type",
                "weights",
                "weight_capacity",
                "penalty",
                "n_items",
                "method",
                "simplifications",
                "n_bins",
                "min_bins",
                "solution",
            ],
            "portfolio_optimization": [
                "problem_type",
                "mu",
                "sigma",
                "risk_factor",
                "penalty",
                "num_assets",
                "budget",
            ],
            "sherrington_kirkpatrick": ["problem_type", "G"],
            "k_color": ["problem_type", "G", "k", "penalty"],
            "generic_qubo": ["problem_type"],
        }

        for k, v in qubos.items():
            assert (
                list(v.problem_instance.keys()) == expected_keys[k]
            ), "Problem instance keys are not correct for problem type {}".format(k)
            assert (
                k == v.problem_instance["problem_type"]
            ), "Problem type is not correct for problem type {}".format(k)

    def test_problem_from_instance_dict(self):
        """
        Test problem from instance method of the problem class.
        """
        problems, qubos = self.__generate_random_problems()

        for type in qubos:
            if type == "generic_qubo":
                continue

            problem_instance = qubos[type].problem_instance.copy()

            problem = create_problem_from_dict(problem_instance)

            assert (
                problem.problem_instance == problems[type].problem_instance
            ), "Problem from instance method is not correct for problem type {}".format(
                type
            )
            assert convert2serialize(problem) == convert2serialize(
                problems[type]
            ), "Problem from instance method is not correct for problem type {}".format(
                type
            )

    def test_qubo_from_dict(self):
        """
        Test qubo from dict method of the QUBO class.
        """

        _, qubos = self.__generate_random_problems()
        for _, qubo in qubos.items():
            qubo_dict = qubo.asdict()

            new_qubo = QUBO.from_dict(qubo_dict)

            for term, new_term in zip(qubo.terms, new_qubo.terms):
                assert set(term) == set(
                    new_term
                ), "QUBO from dict method is not correct for problem type {}, terms compared: {}, {}".format(
                    qubo.problem_instance["problem_type"], term, new_term
                )

                assert set(qubo.weights) == set(
                    new_qubo.weights
                ), "QUBO from dict method is not correct for problem type {}".format(
                    qubo.problem_instance["problem_type"]
                )

                for key in qubo.__dict__:
                    if key != "terms" and key != "weights":
                        assert (
                            qubo.__dict__[key] == new_qubo.__dict__[key]
                        ), "QUBO from dict method is not correct for problem type {}".format(
                            qubo.problem_instance["problem_type"]
                        )

    def test_qubo_terms_and_weight_same_size(self):
        """
        Test that creating a QUBO problem with invalid terms and weights
        sizes raises an exception and check the constant is detected correctly
        """
        n = 2
        terms_wrong = [[0], [1]]
        terms_correct = [[0], [1], []]
        weights = [1, 2, 3]

        with self.assertRaises(ValueError):
            qubo_problem = QUBO(n, terms_wrong, weights)

        qubo_problem = QUBO(n, terms_correct, weights)
        self.assertEqual(qubo_problem.constant, 3)

    def test_qubo_cleaning_terms(self):
        """Test that cleaning terms works for a QUBO problem"""
        terms = [[1, 2], [0], [2, 3], [2, 1], [0]]
        weights = [3, 4, -3, -2, -1]

        cleaned_terms = [[1, 2], [0], [2, 3]]
        self.assertEqual(QUBO.clean_terms_and_weights(terms, weights)[0], cleaned_terms)

    def test_qubo_cleaning_weights(self):
        """Test that cleaning weights works for a QUBO problem"""
        terms = [[1, 2], [0], [2, 3], [2, 1], [0]]
        weights = [3, 4, -3, -2, -1]

        cleaned_weights = [1, 3, -3]
        self.assertEqual(
            QUBO.clean_terms_and_weights(terms, weights)[1], cleaned_weights
        )

    def test_qubo_ising_conversion(self):
        """Test that conversion to Ising formulation works for a QUBO problem"""
        # Small instance
        n = 2
        terms = [[1, 1], [0, 0]]
        weights = [3, 4]

        expected_ising_terms = [[0], [1], []]
        expected_ising_weights = [-2, -1.5, 3.5]

        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(n, terms, weights)
        self.assertEqual(expected_ising_terms, ising_terms)
        self.assertEqual(expected_ising_weights, ising_weights)

        # Larger instance
        n = 4
        terms = [[1, 2], [0], [2, 3], [2, 1], [0]]
        weights = [3, 4, -3, -2, -1]

        expected_ising_terms = [[1, 2], [2, 3], [2, 1], [0], [1], [2], [3], []]
        expected_ising_weights = [0.75, -0.75, -0.5, -1.5, -0.25, 0.5, 0.75, 1.0]

        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(n, terms, weights)
        self.assertEqual(expected_ising_terms, ising_terms)
        self.assertEqual(expected_ising_weights, ising_weights)

    def test_qubo_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # n type-check
        n_list = [1.5, "test", [], (), {}, np.array(1)]
        terms = [[0], []]
        weights = [1, 2]

        with self.assertRaises(TypeError) as e:
            for each_n in n_list:
                QUBO(each_n, terms, weights)
            self.assertEqual(
                "The input parameter, n, has to be of type int", str(e.exception)
            )

        n_list = [-1, 0]
        with self.assertRaises(TypeError) as e:
            for each_n in n_list:
                QUBO(each_n, terms, weights)
            self.assertEqual(
                "The input parameter, n, must be a positive integer greater than 0",
                str(e.exception),
            )

        # weights type-check
        n = 1
        terms = [[0], []]
        weights_list = [{"test": "oh", "test1": "oh"}, np.array([1, 2])]

        for each_weights in weights_list:
            with self.assertRaises(TypeError) as e:
                QUBO(n, terms, each_weights)
            self.assertEqual(
                "The input parameter weights must be of type of list or tuple",
                str(e.exception),
            )

        weights_list = [["test", "oh"], [np.array(1), np.array(2)]]
        for each_weights in weights_list:
            with self.assertRaises(TypeError) as e:
                QUBO(n, terms, each_weights)
            self.assertEqual(
                "The elements in weights list must be of type float or int.",
                str(e.exception),
            )

        # terms type-check
        n = 1
        terms_list = [{"test": [0], "test1": []}]
        weights = [1, 2]
        for each_terms in terms_list:
            with self.assertRaises(TypeError) as e:
                QUBO(n, each_terms, weights)
            self.assertEqual(
                "The input parameter terms must be of type of list or tuple",
                str(e.exception),
            )

    def test_qubo_metadata(self):
        """Test that metadata is correctly stored"""
        qubo_problem = QUBO.random_instance(3)
        qubo_problem.set_metadata({"tag1": "value1", "tag2": "value2"})
        qubo_problem.set_metadata({"tag2": "value2.0"})

        assert (
            qubo_problem.metadata["tag1"] == "value1"
        ), "qubo metadata is not well set"
        assert (
            qubo_problem.metadata["tag2"] == "value2.0"
        ), "qubo metadata is not well set, should have overwritten previous value"

        error = False
        try:
            qubo_problem.set_metadata({"tag10": complex(1, 2)})
        except:
            error = True
        assert (
            error
        ), "Should have thrown an error when setting metadata that is not json serializable"

        error = False
        try:
            qubo_problem.set_metadata({(1, 2): "value"})
        except:
            error = True
        assert (
            error
        ), "Should have thrown an error when setting key metadata that is not json serializable"

    def test_qubo_problem_instance_serializable(self):
        """test that when problem instance is not serializable, it throws an error"""

        qubo = QUBO.random_instance(3)

        error = False
        try:
            qubo.problem_instance = {"tag10": complex(1, 2)}
        except:
            error = True
        assert (
            error
        ), "Should have thrown an error when setting qubo problem instance that is not json serializable"

        error = False
        try:
            qubo.problem_instance = {(1, 2): "value"}
        except:
            error = True
        assert (
            error
        ), "Should have thrown an error when setting key qubo problem instance that is not json serializable"


if __name__ == "__main__":
    unittest.main()
